"""
A collection of analytical expressions of the CGF of privacy RV for popular randomized algorithms
Returns the evaluation.
These functions are used to
"""
import numpy as np

import dpacct


#  CGF of Privacy R.V. of order lambda is equivalent to 1/lambda of the Renyi Divergence of order lambda +1
#  Namely: CGF(lambda) * lambda   =  RenyiDivergence(lambda+1)

# Question: shall we make this symbolic so we can take derivatives?
def CGF_gaussian(params, lamb):
    sigma = params['sigma']
    # assert(sigma >= 0)
    # assert(lamb >= 0)
    return 0.5 / sigma ** 2 * (lamb + 1) * lamb


def CGF_laplace(params, lamb):
    # the parameter is the ratio of the scale parameter and L1 sensitivity
    b = params['b']
    if lamb == 0:
        return 0.0 #(1 / b + np.exp(-1 / b) - 1) * lamb
    elif np.isinf(lamb):
        return 1/b
    else:  # lamb > 0
        return dpacct.stable_logsumexp_two(lamb / b + np.log((lamb + 1.0) / (2 * lamb + 1)), -(lamb + 1.0) / b + np.log(1.0 * lamb / (2 * lamb + 1)))

#            np.log((lamb + 1) / (2 * lamb + 1) * np.exp(lamb / b) + lamb / (2 * lamb + 1) * np.exp(-(lamb + 1) / b))


def CGF_randresponse(params, lamb):
    # the parameter is the bernoulli probability p
    p = params['p']
    if lamb == 0:
        return 0.0 #(2 * p - 1) * np.log(p / (1 - p))
    elif np.isinf(lamb):
        return np.abs(np.log((p/(1-p))))
    else:  # lamb > 0
        return dpacct.stable_logsumexp_two((lamb + 1) * np.log(p) - lamb * np.log(1 - p), (lamb + 1) * np.log(1 - p) - lamb * np.log(p))
            #lamb*np.log(p/(1-p))+np.log(1-p),-lamb*np.log(p/(1-p))+np.log(p))

            #np.log((p / (1 - p)) ** lamb * (1 - p) + (p / (1 - p)) ** (-lamb) * p)


def CGF_diag_gaussian(params, lamb):
    # the parameter is mu1, mu2, sigma1, sigma2
    # they are all d-dimensional numpy arrays
    # Everything can be generalized to general covariance, but
    # calculating A and checking the positive semidefiniteness of A is tricky
    mu1 = params['mu1']
    mu2 = params['mu2']
    sigma1 = params['sigma1']
    sigma2 = params['sigma2']

    def extrapolate(a, b):
        return (lamb + 1) * a - lamb * b

    A = extrapolate(sigma1 ** (-1), sigma2 ** (-1))
    if not (A > 0).all():
        return np.inf
    else:
        sigma = extrapolate(sigma2, sigma1)
        tmp=(lamb * (lamb + 1) / 2 * np.dot((mu1 - mu2), sigma * (mu1 - mu2)) - 0.5 * (np.sum(np.log(sigma))
                                                                                   - extrapolate(
            np.sum(np.log(sigma2)), np.sum(np.log(sigma1)))))
        #

        return (lamb * (lamb + 1) / 2 * np.dot((mu1 - mu2), sigma * (mu1 - mu2)) - 0.5 * (np.sum(np.log(sigma))
                                                                                          - extrapolate(
            np.sum(np.log(sigma2)), np.sum(np.log(sigma1)))))
        #
        # tmp1= mu1/sigma1
        # tmp2 = mu2/sigma2
        # tmp = extrapolate(tmp1,tmp2)
        # mag1= np.dot(mu1,tmp1)
        # mag2 =np.dot(mu2,tmp2)
        # return (0.5*np.sum(np.log(sigma2)) - 0.5*np.sum(np.log(sigma2)) + 1/2/(lamb)*(-np.sum(A)-np.sum(sigma1))
        #         - (extrapolate(mag1,mag2) - np.dot(tmp,tmp/A))/(2*lamb)) * lamb

def CGF_asymp_subsampled_gaussian(params, lamb):
    sigma = params['sigma']
    prob = params['prob']
    assert(prob<1 and prob >= 0)
    return prob**2/(1-prob)**4 / sigma**2 *lamb *(lamb+1)


def CGF_expfamily(params, lamb):
    eta1 = params['eta1']
    eta2 = params['eta2']  # these are the worst natural parameters of the exp family distributions.
    A = params['A']  # A is a lambda function handle the log partition. A needs to handle cases of infinity

    def extrapolate(a, b):
        return (lamb + 1) * a - lamb * b

    return A(extrapolate(eta1, eta2)) - extrapolate(A(eta1), A(eta2))