# This is the main file of the python package dp-acct
# that keeps track of the privacy loss of a sequence of randomized algorithms.


import numpy as np
from scipy.optimize import fmin_bfgs,minimize_scalar
import math
# utiltiy function for calculating the combinatorials
# save log(x+1).

def stable_logsumexp(x):
    a = np.max(x)
    return a+np.log(np.sum(np.exp(x-a)))
def stable_logsumexp_two(x,y):
    a = np.maximum(x,y)
    return a + np.log(np.exp(x-a) + np.exp(y-a))


def get_binom_coeffs(sz):
    C = np.zeros(shape = (sz + 1, sz + 1));
    #for k in range(1,sz + 1,1):
    #    C[0, k] = -np.inf
    for n in range(sz + 1):
        C[n, 0] = 0  # 1
    for n in range(1,sz + 1,1):
        C[n, n] = 0
    for n in range(1,sz + 1,1):
        for k in range(1,n,1):
            # numerical stable way of implementing the recursion rule
            C[n, k] = stable_logsumexp_two(C[n - 1, k - 1],C[n - 1, k])
    # only the lower triangular part of the matrix matters
    return C

def get_binom_coeffs_dict(sz):
    C = {} #np.zeros(shape = (sz + 1, sz + 1));
    #for k in range(1,sz + 1,1):
    #    C[0, k] = -np.inf
    for n in range(sz + 1):
        C[(n, 0)] = 0  # 1
    for n in range(1,sz + 1,1):
        C[(n, n)] = 0
    for n in range(1,sz + 1,1):
        for k in range(1,n,1):
            # numerical stable way of implementing the recursion rule
            C[(n, k)] = stable_logsumexp_two(C[(n - 1, k - 1)],C[(n - 1, k)])
    # only the lower triangular part of the matrix matters
    return C

def expand_binom_coeffs_dict(C,sz,sznew):
    for n in range(sz,sznew+1,1):
        C[(n,0)] = 0
    for n in range(sz,sznew + 1,1):
        C[(n, n)] = 0
    for n in range(sz,sznew + 1,1):
        for k in range(1,n,1):
            C[(n, k)] = stable_logsumexp_two(C[(n - 1, k - 1)],C[(n - 1, k)])
    return C # also need to update the size of C to sznew whenever this function is called just to keep track.

# Get the eps and delta for a single Gaussian mechanism
def get_eps_gaussian(sigma, delta):
    """ This function calculates the eps for Gaussian Mech given a delta"""

    lamb = (np.log(1 / delta)*2)**0.5 * sigma

    # The following block solves for lamb numerically, in case it is not analytically tractable
    #def F(lamb):
    #    return (np.log(1 / delta) + 0.5 / sigma ** 2 * lamb * (lamb + 1)) / lamb
    #def G(x):
    #    return -np.log(1 / delta)/ x **2 +  0.5 / sigma ** 2
    #lamb = fmin_bfgs(F, 5, G, disp= True)
    return (np.log(1 / delta) + 0.5 / sigma ** 2 * lamb*(lamb+1)) / lamb


def RDP_via_CGF(CGF):
    """ Takes in the lambda function CGF, convert it to a lambda function that does RDP"""
    g = lambda alpha: (alpha - 1)*CGF(alpha-1)
    return g


def subsample_epsdelta(eps,delta,prob):
    # Theorem 1 of :
    # Li, Ninghui, Qardaji, Wahbeh, and Su, Dong.
    # On sampling, anonymization, and differential privacy or, k-anonymization meets differential privacy
    return np.log(1+prob*(np.exp(eps)-1)), prob*delta


class DPAcct:
    """A class that keeps track of (eps,delta) of all mechanisms that got run so far"""
    #
    #DPlosses = []

    def __init__(self,disp=False):
        self.DPlosses = []
        self.eps_state1 = 0
        self.eps_state2 = 0
        self.eps_state3 = 0
        self.delta_state = 0
        self.delta_state2 = 0
        self.disp=disp

    def update_DPlosses(self,eps,delta):
        self.DPlosses.append([eps,delta])
        self.eps_state1 += eps
        self.eps_state2 += (np.exp(eps) - 1) * eps / (np.exp(eps) + 1)
        self.eps_state3 += eps ** 2
        self.delta_state += np.log(1-delta)
        self.delta_state2 += delta
        #update the optimal DPlosses here?

    def get_eps_delta_naive(self):
        return self.eps_state1, self.delta_state2

    def get_minimum_possible_delta(self):
        return 1-np.exp(self.delta_state)


    def get_eps(self,delta):
        """ make use of KOV'15 to calculate the composition for hetereogeneous mechanisms"""
        # TODO: further improve upon this with Salil Vadhan's approximation algorithm
        if delta>1 or delta<0:
            if self.disp:
                print("The chosen delta is clearly not feasible.")
            return -1
        if delta is 0:
            if self.delta_state2 is 0:
                return self.delta_state2
            else:
                return np.inf
                #if self.disp:
                #    print("Pure DP is not feabile. Choose non-zero delta")
                #return -1

        # 1-delta =  (1-deltatilde) exp(state)
        deltatilde = 1 - np.exp(np.log(1-delta) -self.delta_state)
        if deltatilde <= 0:
            return np.inf
            #if self.disp:
            #    print("The chosen delta is not feasible. delta needs to be at least ", 1-np.exp(self.delta_state))
            #return -1

        eps1 = self.eps_state1
        eps2 = self.eps_state2 + (self.eps_state3 *2 *  np.log(np.exp(1) + self.eps_state3 ** 0.5 / deltatilde ))**0.5
        eps3 = self.eps_state2 + (2*self.eps_state3*np.log( 1/ deltatilde))**0.5

        return np.minimum(np.minimum(eps1,eps2),eps3)




class anaCGFAcct:
    """A class that keeps track of the analytical expression of the CGF of the privacy loss R.B."""
    def __init__(self,m=10,tol=0.1,m_max=1000):
        self.m = m # default number of binomial coefficients to precompute
        self.m_max = m_max
        self.lambs = np.linspace(1, self.m, self.m).astype(int)
        self.CGFs_int = np.zeros_like(self.lambs,float)

        self.n=0
        self.CGFs = [] # analytical CGFs
        self.coeffs = []
        self.CGF_inf = .0 # This is effectively for pure DP.
        self.logBinomC = get_binom_coeffs(self.m + 1)  # The logBinomC is only needed for subsampling mechanisms.
        self.idxhash = {} # save the index of previously used algorithms
        self.cache = {} # dictionary to save results from previously seen algorithms
        self.evalCGF = lambda x: 0
        self.flag = True # a flag indicating whether evalCGF is out of date
        self.flag_subsample = False # a flag to indicate whether we need to expand the logBinomC.
        self.tol = tol

    def build_zeroth_oracle(self):
        self.evalCGF = lambda x:  sum([c*item(x) for (c,item) in zip(self.coeffs, self.CGFs)])

    def plot_cgf(self):
        if not self.flag:
            self.build_zeroth_oracle()
            self.flag = True

        import matplotlib.pyplot as plt
        plt.figure(num=None, figsize=(12, 8), dpi=80, facecolor='w', edgecolor='k')
        x = range(0,self.m,1)
        y = [ self.evalCGF(item) for item in x]
        plt.plot(x, y)
        plt.show()


    def plot_cgf_int(self):
        import matplotlib.pyplot as plt
        plt.figure(num=None, figsize=(12, 8), dpi=80, facecolor='w', edgecolor='k')
        plt.plot(self.lambs, self.CGFs_int)
        plt.show()

    def get_eps(self, delta): # minimize over \lambda
        if not self.flag:
            self.build_zeroth_oracle()
            self.flag = True

        if delta<0 or delta > 1:
            print("Error! delta is a probability and must be between 0 and 1")
        if delta is 0:
            return self.CGF_inf
        else:
            def fun(x):
                if x == 0:
                    return np.inf
                else:
                    return (np.log(1 / delta) + self.evalCGF(x)) / x

            def fun_int(i):
                return (np.log(1 / delta) + self.CGFs_int[i-1]) / i

            #fun  = lambda x: (np.log(1/delta) + self.evalCGF(x)) / x
            # First check if the forward difference is positive at self.m, or if it is infinite
            while (self.m<self.m_max) and (not np.isposinf(fun(self.m))) and (fun_int(self.m-1)-fun_int(self.m-2) < 0):
                #print([fun_int(self.m-1),fun_int(self.m-2),fun_int(self.m-1)-fun_int(self.m-2)])
                # If so, double m, expand logBimomC until it is positive


                if self.flag_subsample:
                    self.logBinomC = get_binom_coeffs(self.m*2+1)
                    #expand_binom_coeffs_dict(self.logBinomC,self.m+1,self.m*2+1)

                # update the integer CGF and the cache for each function
                cgf_int_new = np.zeros_like(self.CGFs_int)
                for key,val in self.cache.items():
                    if type(key) is tuple:
                        func = key[0]
                    else:
                        func = key
                    idx = self.idxhash[key]
                    cgf = self.CGFs[idx]
                    newarray = np.zeros_like(self.CGFs_int)
                    for j in range(self.m):
                        newarray[j] = cgf(j+self.m+1)

                    # This following line is very slow. Do not use it.
                    #newarray = np.array([cgf(j) for j in range(self.m + 1, self.m*2+1,1)])

                    coeff = self.coeffs[idx]
                    cgf_int_new += newarray * coeff
                    self.cache[key] = np.concatenate((val, newarray))
                # update the corresponding quantities
                newlambs = range(self.m + 1, self.m*2+1,1)
                self.CGFs_int = np.concatenate((self.CGFs_int, cgf_int_new))
                self.lambs = np.concatenate((self.lambs, np.array(newlambs))) # array of integers
                self.m = self.m*2


            # while (not np.isposinf(fun(self.m))) and (fun(self.m) - fun(self.m-1) < 0):
            #     print(fun(self.m) - fun(self.m - 1))
            #     # If so, double m, expand logBimomC until it is positive
            #     if self.flag_subsample:
            #         self.logBinomC = get_binom_coeffs(self.m*2+1)
            #         #expand_binom_coeffs_dict(self.logBinomC,self.m+1,self.m*2+1)
            #     self.m = self.m*2

            # Lastly, solve the local binary search problem to identify the solution. Return the optimal value.

            bestint = np.argmin((np.log(1 / delta) + self.CGFs_int) / self.lambs)

            #self.plot_cgf_int()
            #results = minimize_scalar(fun, tol=0.01, bounds=[0,self.m])
            if bestint == self.m-1:
                return fun(bestint)

            results = minimize_scalar(fun, tol=self.tol,method='Bounded',bounds=[bestint, bestint+2],
                                      options={'disp':False})
            # the while loop above ensures that bestint+2 is at most m, and also bestint is at least 0.
            if results.success:
                return results.fun
            else:
                # There are cases when certain \delta is not feasible.
                # For example, let p and q be uniform the privacy R.V. is either 0 or \infty and unless all \infty
                # events are taken cared of by \delta, \epsilon cannot be < \infty
                return -1

    def compose_mechanism(self, func):
        self.flag = False
        if func in self.idxhash:
            self.coeffs[self.idxhash[func]] += 1
            # also update the integer CGFs
            self.CGFs_int += self.cache[func]
        else:
            # book keeping
            self.idxhash[func] = self.n
            self.n += 1
            self.coeffs.append(1.0)
            # update the analytical
            self.CGFs.append(func)

            # also update the integer results
            if func in self.cache:
                tmp = self.cache[func]
            else:
                tmp = np.zeros_like(self.CGFs_int, float)
                for i in range(self.m):
                    tmp[i] = func(i+1)
                self.cache[func] = tmp  # save in cache
            self.CGFs_int += tmp
            self.CGF_inf += func(np.inf)

    def compose_subsampled_mechanism(self,func, prob):
        self.flag = False
        self.flag_subsample = True
        if (func, prob) in self.idxhash:
            idx = self.idxhash[(func, prob)]
            # update the coefficients of each function
            self.coeffs[idx] += 1
            # also update the integer CGFs
            self.CGFs_int += self.cache[(func, prob)]
        else:
            def subsample_func_int(x):
                if np.isinf(func(x)):
                    return np.inf
                moments = [np.log(2)+(j+1)*np.log(prob)+self.logBinomC[int(x+1), j+1]+ func(j) for j in range(1,int(x+1),1)]
                #moments = [np.log(2) + (j + 1) * np.log(prob) + self.logBinomC[int(x + 1), j + 1] + func(j) for j in
                #           range(1, int(x + 1), 1)]
                tmp = [0]+moments
                return stable_logsumexp([0]+moments)

            def subsample_func(x): # linear interpolation upper bound
                if np.isinf(x):
                    return
                if np.equal(np.mod(x, 1), 0):
                    return subsample_func_int(x)
                xc = math.ceil(x)
                xf = math.floor(x)
                return (x-xf)*subsample_func_int(xc) + (1-(x-xf))*subsample_func_int(xf)

            # book keeping
            self.idxhash[(func, prob)] = self.n # save the index
            self.n += 1 # increment the number of unique mechanisms
            self.coeffs.append(1.0) # Update the coefficient
            self.CGFs.append(subsample_func) # update the analytical functions

            # also update the integer results
            if (func,prob) in self.cache:
                results = self.cache[(func,prob)]
            else:
                results = np.zeros_like(self.CGFs_int, float)
                m = np.max(self.lambs)
                jvec = np.arange(2, m + 2)
                logterm3plus = np.zeros_like(results)
                for j in jvec:
                    logterm3plus[j-2] = np.log(2) + j * (np.log(prob)) + func(j-1)
                for lamb in range(2, m + 1):
                    if np.isinf(logterm3plus[lamb-1]):
                        results[lamb-1] = np.inf
                    else:
                        tmp = stable_logsumexp(logterm3plus[0:lamb] + self.logBinomC[lamb + 1, 2:(lamb + 2)])
                        results[lamb - 1] = stable_logsumexp_two(0, tmp)
                self.cache[(func,prob)] = results # save in cache
            self.CGFs_int += results

            eps, delta = subsample_epsdelta(func(np.inf), 0, prob)
            self.CGF_inf += eps







class CGFAcct:
    """A class that keep tracks of the cumulative generating functions of the privacy loss R.V."""
    # input to this algorithm is a sequence of lambdas we want to keep track of.
    # 1-100 sounds like a reasonable range that we want to keep track of


    def __init__(self,m):
        self.m=m
        self.lambs = np.linspace(1, self.m, self.m).astype(int)
        self.CGFs = np.zeros_like(self.lambs,float)
        self.CGF_inf = .0
        self.logBinomC = get_binom_coeffs(self.m + 1)
        self.cache = dict()
        # Interface for getting



    def get_eps(self, delta): # find the smallest eps that corresponds to a delta through a tail bound
        #print(self.CGFs)
        if delta<0 or delta > 1:
            print("Error! delta is a probability and must be between 0 and 1")
            return -1
        if delta is 0:
            return self.CGF_inf
        else:
            #print(np.argmin((np.log(1 / delta) + self.CGFs) / self.lambs))
            return np.min((np.log(1 / delta) + self.CGFs) / self.lambs)




    # How about precomputing it?
    def update_cgf_subsamplegaussian(self, prob, sigma):
        """ update CGF for a subsampleGaussian mechanism """
        # This is for probability of sampling and then adding gaussian noise
        # with a normalized noise level of sigma
        # prob needs to be a floating point number in (0,1)

        # The first term is 1, the second term is prob^2*lamb*(lamb+1)
        # The third term onwards up to lamb+1 term is
        # {lamb+1\choose j} * prob^j[1-e^{-sigma^2/2}]^je^{(j^2+1)sigma^2/2}

        # term2 = prob ** 2 * (np.exp(sigma ** 2) - np.exp(-sigma ** 2));

        if (prob,sigma) in self.cache:
            self.CGFs += self.cache[(prob, sigma)]
            CGF_inf = np.inf
            return

        m = np.max(self.lambs)
        j = np.arange(2, m+2)
        logterm3plus = np.log(2) + j * (np.log(prob)) + 1.0*j*(j-1)/(2*sigma**2)
        # make sure that lambs is n dimensional.
        results = np.zeros_like(self.lambs,float)

        for lamb in range(2, m+1):
            tmp = stable_logsumexp(logterm3plus[0:lamb] + self.logBinomC[lamb + 1, 2:(lamb + 2)])
            results[lamb-1] = stable_logsumexp_two(0, tmp)

        # cache the calculations so there's no need to do it again..
        self.cache[(prob,sigma)] = results

        self.CGFs += results
        self.CGF_inf = np.inf

    def update_cgf_gaussian(self,sigma):
        """ update CGF for a Gaussian mechanism """
        results = 0.5/sigma**2 * (self.lambs+1) * self.lambs

        self.CGFs += results
        self.CGF_inf = np.inf
        # in this cae everything can be calculated exactly

    def update_cgf_laplace(self,sigma):
        """ update CGF for a Laplace mechanism """
        # sigma is the lambda of Laplace distribution
        results = np.zeros_like(self.CGFs)
        results[0] = 1/sigma + np.exp(-1/sigma)-1
        results[1:] = 1/self.lambs[1:] * np.log((self.lambs[1:]+1)/(2*self.lambs[1:]+1)*np.exp(self.lambs[1:]/sigma)
                                                + self.lambs[1:]/(2*self.lambs[1:]+1)*np.exp(-(self.lambs[1:]+1)/sigma))
        results[:] = results*self.lambs
        self.CGFs += results
        self.CGF_inf += 1/sigma

#    def update_cgf_exponential(self,sigma):
#        """ update CGF for a Exponential mechanism with an insensitive cost function"""
#        print("Not implemented yet\n")

    def update_cgf_randresponse(self,p):
        """ update CGF for a Randomized response of {0,1} variable """
        # Randomized response toggle the response with probability p
        results = np.zeros_like(self.CGFs)
        results[0] = (2*p-1)*np.log(p/(1-p))
        results[1:] = 1/self.lambs[1:] * stable_logsumexp_two((self.lambs[1:]+1)*np.log(p)-self.lambs[1:]*np.log(1-p),
                                                              (self.lambs[1:]+1)*np.log(1-p) - self.lambs[1:]*np.log(p))
        results = results * self.lambs
        self.CGFs += results
        self.CGF_inf += np.abs(np.log((p/(1-p))))


# TODO: Getting rid of the dependence on parameter m
# TODO: Do an accounting of analytical functions of these moments by solve a 1D nonlinear optimization each time  we query for \eps with a fixed \delta.
# TODO:


